# Credit some part of the source code from bilby

import jax.numpy as jnp
from jaxgw.PE.constants import *


def make_detector_response(detector_tensor, detector_vertex):
    antenna_response_plus = make_antenna_response(detector_tensor,'plus')
    antenna_response_cross = make_antenna_response(detector_tensor, 'cross')
    def detector_response(f, hp, hc, ra, dec, time, psi):
        output = antenna_response_plus(ra, dec, time, psi)*hp + antenna_response_cross(ra, dec, time, psi)*hc
        timeshift = time_delay_geocentric(detector_vertex, jnp.array([0.,0.,0.]), ra, dec, time)
        output = output * jnp.exp(-1j * 2 * jnp.pi * f * timeshift)
        return output
    return detector_response
        
##########################################################
# Construction of arms
##########################################################

def construct_arm(latitude, longitude, arm_tilt, arm_azimuth):
    """

     Args:

        latitude: Latitude in radian
        longitude: Longitude in radian
        arm_tilt: Arm tilt in radian
        arm_azimuth: Arm azimuth in radian
   
    """

    e_long = jnp.array([-jnp.sin(longitude), jnp.cos(longitude), 0])
    e_lat = jnp.array([-jnp.sin(latitude) * jnp.cos(longitude),
                      -jnp.sin(latitude) * jnp.sin(longitude), jnp.cos(latitude)])
    e_h = jnp.array([jnp.cos(latitude) * jnp.cos(longitude),
                    jnp.cos(latitude) * jnp.sin(longitude), jnp.sin(latitude)])

    return (jnp.cos(arm_tilt) * jnp.cos(arm_azimuth) * e_long +
            jnp.cos(arm_tilt) * jnp.sin(arm_azimuth) * e_lat +
            jnp.sin(arm_tilt) * e_h)


def detector_tensor(arm1, arm2):
    return 0.5 * (jnp.einsum('i,j->ij', arm1, arm1) - jnp.einsum('i,j->ij', arm2, arm2))

##########################################################
# Construction of detector tensor
##########################################################

def make_get_polarization_tensor(mode):

    """

    Since most of the application will only use specific modes,
    this function hoist the if-else loop out from the actual kernel to save time from compiling the kernel.

    Args:
        mode: string

    """

    if mode.lower() == 'plus':
        kernel = lambda m,n: jnp.einsum('i,j->ij', m, m) - jnp.einsum('i,j->ij', n, n)
    elif mode.lower() == 'cross':
        kernel = lambda m,n: jnp.einsum('i,j->ij', m, n) + jnp.einsum('i,j->ij', n, m)
    elif mode.lower() == 'breathing':
        kernel = lambda m,n: jnp.einsum('i,j->ij', m, m) + jnp.einsum('i,j->ij', n, n)

        # Calculating omega here to avoid calculation when model in [plus, cross, breathing]
        if mode.lower() == 'longitudinal':
            def kernel(m,n):
                omega = jnp.cross(m, n)
                return jnp.einsum('i,j->ij', omega, omega)
        elif mode.lower() == 'x':
            def kernel(m,n):
                omega = jnp.cross(m, n)
                return jnp.einsum('i,j->ij', m, omega) + jnp.einsum('i,j->ij', omega, m)
        elif mode.lower() == 'y':
            def kernel(m,n):
                omega = jnp.cross(m, n)
                return jnp.einsum('i,j->ij', n, omega) + jnp.einsum('i,j->ij', omega, n)
        else:
            raise ValueError("{} not a polarization mode!".format(mode))

    def get_polarization_tensor(ra, dec, time, psi):
        gmst = jnp.mod(time, 2 * jnp.pi)
        phi = ra - gmst
        theta = jnp.pi / 2 - dec

        u = jnp.array([jnp.cos(phi) * jnp.cos(theta), jnp.cos(theta) * jnp.sin(phi), -jnp.sin(theta)])
        v = jnp.array([-jnp.sin(phi), jnp.cos(phi), 0])
        m = -u * jnp.sin(psi) - v * jnp.cos(psi)
        n = -u * jnp.cos(psi) + v * jnp.sin(psi)

        return kernel(m, n)
    
    return get_polarization_tensor

    
def make_antenna_response(detector_tensor, mode):
    kernel = make_get_polarization_tensor(mode)
    def antenna_response(ra, dec, time, psi):
        polarization_tensor = kernel(ra, dec, time, psi)
        return jnp.einsum('ij,ij->', detector_tensor, polarization_tensor)
    return antenna_response

def time_delay_geocentric(detector1, detector2, ra, dec, time):
    """
    Calculate time delay between two detectors in geocentric coordinates based on XLALArrivaTimeDiff in TimeDelay.c

    Parameters
    ==========
    detector1: array_like
        Cartesian coordinate vector for the first detector in the geocentric frame
        generated by the Interferometer class as self.vertex.
    detector2: array_like
        Cartesian coordinate vector for the second detector in the geocentric frame.
        To get time delay from Earth center, use detector2 = np.array([0,0,0])
    ra: float
        Right ascension of the source in radians
    dec: float
        Declination of the source in radians
    time: float
        GPS time in the geocentric frame

    Returns
    =======
    float: Time delay between the two detectors in the geocentric frame

    """
    gmst = jnp.mod(time, 2 * jnp.pi)
    phi = ra - gmst
    theta = jnp.pi / 2 - dec
    omega = jnp.array([jnp.sin(theta) * jnp.cos(phi), jnp.sin(theta) * jnp.sin(phi), jnp.cos(theta)])
    delta_d = detector2 - detector1
    return jnp.dot(omega, delta_d) / speed_of_light

def get_vertex_position_geocentric(latitude, longitude, elevation):
    """
    Calculate the position of the IFO vertex in geocentric coordinates in meters.

    Based on arXiv:gr-qc/0008066 Eqs. B11-B13 except for the typo in the definition of the local radius.
    See Section 2.1 of LIGO-T980044-10 for the correct expression

    Parameters
    ==========
    latitude: float
        Latitude in radians
    longitude:
        Longitude in radians
    elevation:
        Elevation in meters

    Returns
    =======
    array_like: A 3D representation of the geocentric vertex position

    """
    semi_major_axis = 6378137  # for ellipsoid model of Earth, in m
    semi_minor_axis = 6356752.314  # in m
    radius = semi_major_axis**2 * (semi_major_axis**2 * jnp.cos(latitude)**2 +
                                   semi_minor_axis**2 * jnp.sin(latitude)**2)**(-0.5)
    x_comp = (radius + elevation) * jnp.cos(latitude) * jnp.cos(longitude)
    y_comp = (radius + elevation) * jnp.cos(latitude) * jnp.sin(longitude)
    z_comp = ((semi_minor_axis / semi_major_axis)**2 * radius + elevation) * jnp.sin(latitude)
    return jnp.array([x_comp, y_comp, z_comp])







# Import packages
from typing import List, Tuple
import lalsimulation as lalsim 
import jax.numpy as jnp
import jax
import numpy as np
jax.config.update('jax_enable_x64', True)

psd_func_dict = {
    'H1': lalsim.SimNoisePSDaLIGOZeroDetHighPower,
    'L1': lalsim.SimNoisePSDaLIGOZeroDetHighPower,
    'V1': lalsim.SimNoisePSDAdvVirgo,
}

def generate_noise(seed: int, f_sampling: int = 2048, duration: int = 4, f_min: float = 30., ifos: List = ['H1', 'L1']):


    # define sampling rate and duration

    delta_t = 1/f_sampling
    tlen = int(round(duration / delta_t))

    freqs = np.fft.rfftfreq(tlen, delta_t)
    delta_f = freqs[1] - freqs[0]

    # we will want to pad low frequencies; the function below applies a
    # prescription to do so smoothly, but this is not really needed: you
    # could just set all values below `fmin` to a constant.
    def pad_low_freqs(f, psd_ref):
        return psd_ref + psd_ref*(f_min-f)*jnp.exp(-(f_min-f))/3

    psd_dict = {}
    for ifo in ifos:
        psd = np.zeros(len(freqs))
        for i,f in enumerate(freqs):
            if f >= f_min:
                psd[i] = psd_func_dict[ifo](f)
            else:
                psd[i] = pad_low_freqs(f, psd_func_dict[ifo](f_min))
        psd_dict[ifo] = jnp.array(psd,dtype=jnp.float64)

    rng_key = jax.random.PRNGKey(seed)
    rng_keys = jax.random.split(rng_key)

    noise_fd_dict = {}
    for ifo, psd in psd_dict.items():
        rng_keys = jax.random.split(rng_keys[0], 3)
        var = psd / (4.*delta_f)  # this is the variance of LIGO noise given the definition of the likelihood function
        noise_real = jax.random.normal(rng_keys[1],shape=(len(psd),))*jnp.sqrt(var)
        noise_imag = jax.random.normal(rng_keys[2],shape=(len(psd),))*jnp.sqrt(var)
        noise_fd_dict[ifo] = noise_real + 1j*noise_imag

    return freqs, psd_dict, noise_fd_dict


import jax.numpy as jnp
from jax import jit

@jit
def inner_product(h1, h2, frequency, PSD):
	"""
	Do PSD interpolation outside the inner product loop to speed up the evaluation
	"""
	#psd_interp = jnp.interp(frequency, PSD_frequency, PSD)
	df = frequency[1] - frequency[0]
	integrand = jnp.conj(h1)* h2 / PSD
	return 4. * jnp.real(jnp.trapz(integrand,dx=df))

@jit
def m1m2_to_Mq(m1,m2):
	"""
	Transforming the primary mass m1 and secondary mass m2 to the Total mass M
	and mass ratio q.

	Args:
		m1: Primary mass of the binary.
		m2: Secondary mass of the binary.

	Returns:
		A tuple containing both the total mass M and mass ratio q.
	"""
	M_tot = jnp.log(m1+m2)
	q = jnp.log(m2/m1)-jnp.log(1-m2/m1)
	return M_tot, q

@jit
def Mq_to_m1m2(trans_M_tot,trans_q):
	M_tot = jnp.exp(trans_M_tot)
	q = 1./(1+jnp.exp(-trans_q))
	m1 = M_tot/(1+q)
	m2 = m1*q
	return m1, m2

@jit
def Mc_q_to_m1m2(Mc,q):
	eta = q/(1+q)**2
	M_tot = Mc/eta**(3./5)
	m1 = M_tot/(1+q)
	m2 = m1*q
	return m1, m2

def ra_dec_to_theta_phi(ra, dec, gmst):
    phi = ra - gmst
    theta = jnp.pi / 2 - dec
    return theta, phi




from jax import jit
from jaxgw.PE.detector_projection import get_detector_response
from jaxgw.PE.utils import inner_product

def single_detector_likelihood(waveform_model, params, data, data_f, PSD, detector):
    waveform = waveform_model(data_f, params)
    waveform = get_detector_response(waveform, params, detector)
    match_filter_SNR = inner_product(waveform, data, data_f, PSD)
    optimal_SNR = inner_product(waveform, waveform, data_f, PSD)
    return (-2*match_filter_SNR + optimal_SNR)/2






from lal import MSUN_SI, PC_SI, MTSUN_SI
import lalsimulation as lalsim
import numpy as np

from jaxgw.gw.waveform.IMRPhenomC import IMRPhenomC
from jaxgw.gw.waveform.TaylorF2 import TaylorF2

mass_1 = 30.
mass_2 = 30.
luminosity_distance = 410.
f0 = 20.
max_f = 2048
delta_f = 1./8
spin = 0.

injection_parameters = dict(
	mass_1=mass_1, mass_2=mass_2, spin_1=0.0, spin_2=0.0, luminosity_distance=luminosity_distance, 	phase_c=0, t_c=0, theta_jn=0.4, psi=2.659,)

 
waveform1 = lalsim.SimIMRPhenomCGenerateFD(0., delta_f, mass_1*MSUN_SI, mass_2* MSUN_SI, 0., f0, max_f,luminosity_distance* 1e6*PC_SI,{})
frequency = waveform1.f0 + np.arange(len(waveform1.data.data)) * waveform1.deltaF
waveform2 = IMRPhenomC(frequency,injection_parameters)
waveform3 = TaylorF2(frequency,injection_parameters)


from jaxgw.PE.detector_projection import construct_arm, detector_tensor, get_vertex_position_geocentric
import jax.numpy as jnp

# See https://git.ligo.org/lscsoft/bilby/-/tree/master/bilby/gw/detector/detectors for detector parameters.

degree_to_radian = jnp.pi/180

def get_H1():
	"""
	Get the detector response matrix and the vertex position for H1.

	Returns
	-------
	H1_detector_response : ndarray
		The detector response matrix for H1.
	H1_vertex : ndarray
		The vertex position for H1.
	"""
	H1_lat = (46 + 27. / 60 + 18.528 / 3600) * degree_to_radian
	H1_long = -(119 + 24. / 60 + 27.5657 / 3600) * degree_to_radian
	H1_xarm_azimuth = 125.9994 * degree_to_radian
	H1_yarm_azimuth = 215.9994 * degree_to_radian
	H1_xarm_tilt = -6.195e-4
	H1_yarm_tilt = 1.25e-5
	H1_elevation = 142.554
	
	H1_arm1 = construct_arm(H1_lat, H1_long, H1_xarm_tilt, H1_xarm_azimuth)
	H1_arm2 = construct_arm(H1_lat, H1_long, H1_yarm_tilt, H1_yarm_azimuth)

	H1_vertex = get_vertex_position_geocentric(H1_lat, H1_long, H1_elevation)
	
	return detector_tensor(H1_arm1, H1_arm2), H1_vertex

def get_L1():
	"""
	Get the detector response matrix and the vertex position for L1.

	Returns
	-------
	L1_detector_response : ndarray
		The detector response matrix for L1.
	L1_vertex : ndarray
		The vertex position for L1.
		
	"""
	L1_lat = (30 + 33. / 60 + 46.4196 / 3600) * degree_to_radian
	L1_long = -(90 + 46. / 60 + 27.2654 / 3600) * degree_to_radian
	L1_xarm_azimuth = 197.7165 * degree_to_radian
	L1_yarm_azimuth = 287.7165 * degree_to_radian
	L1_xarm_tilt = 0 
	L1_yarm_tilt = 0
	L1_elevation = -6.574
	
	L1_arm1 = construct_arm(L1_lat, L1_long, L1_xarm_tilt, L1_xarm_azimuth)
	L1_arm2 = construct_arm(L1_lat, L1_long, L1_yarm_tilt, L1_yarm_azimuth)

	L1_vertex = get_vertex_position_geocentric(L1_lat, L1_long, L1_elevation)
	
	return detector_tensor(L1_arm1, L1_arm2), L1_vertex

def get_V1():
	"""
	Get the detector response matrix and the vertex position for V1.

	Returns
	-------
	V1_detector_response : ndarray
		The detector response matrix for V1.
	V1_vertex : ndarray
		The vertex position for V1.
	"""
	V1_lat = (43 + 37. / 60 + 53.0921 / 3600) * degree_to_radian
	V1_long = (10 + 30. / 60 + 16.1878 / 3600) * degree_to_radian
	V1_xarm_azimuth = 70.5674 * degree_to_radian
	V1_yarm_azimuth = 160.5674 * degree_to_radian
	V1_xarm_tilt = 0
	V1_yarm_tilt = 0
	V1_elevation = 51.884
	
	V1_arm1 = construct_arm(V1_lat, V1_long, V1_xarm_tilt, V1_xarm_azimuth)
	V1_arm2 = construct_arm(V1_lat, V1_long, V1_yarm_tilt, V1_yarm_azimuth)

	V1_vertex = get_vertex_position_geocentric(V1_lat, V1_long, V1_elevation)
	
	return detector_tensor(V1_arm1, V1_arm2), V1_vertex



from lal import MSUN_SI, PC_SI, MTSUN_SI
import lalsimulation as lalsim
import numpy as np

from jaxgw.gw.waveform.TaylorF2 import TaylorF2
from bilby.gw.utils import greenwich_mean_sidereal_time

mass_1 = 30.
mass_2 = 30.
luminosity_distance = 410.
f0 = 20.
max_f = 2048
delta_f = 1./8
spin = 0.02

injection_parameters = dict(
	mass_1=mass_1, mass_2=mass_2, spin_1=spin, spin_2=spin, luminosity_distance=luminosity_distance, phase_c=0, t_c=0,\
	theta_jn=0.4, psi=2.659,f_ref = 50)

 
waveform1 = lalsim.SimInspiralTaylorF2(0., delta_f, mass_1*MSUN_SI, mass_2* MSUN_SI, 0., 0., f0, max_f, 50,luminosity_distance* 1e6*PC_SI,{})
waveform2 = lalsim.SimInspiralChooseFDWaveform(mass_1*MSUN_SI, mass_2*MSUN_SI, 0, 0, 0, 0, 0, 0, luminosity_distance*1e6*PC_SI,0.4,0,0,0,0,1./8,40,2048,50,{},5)
frequency = waveform1.f0 + np.arange(len(waveform1.data.data)) * waveform1.deltaF
waveform3 = TaylorF2(frequency,injection_parameters)

import bilby

duration = 32
sampling_frequency = 2 * 1024

# Fixed arguments passed into the source model. The analysis starts at 40 Hz.
waveform_arguments = dict(waveform_approximant='TaylorF2',
                          reference_frequency=50., minimum_frequency=40.0)

# Create the waveform_generator using a LAL Binary Neutron Star source function
waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration, sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.lal_binary_neutron_star,
    parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_neutron_star_parameters,
    waveform_arguments=waveform_arguments)

waveform4 = waveform_generator.frequency_domain_source_model(frequency, mass_1, mass_2, luminosity_distance, 0, 0, 0, 0, 0, 0, 0.4, 0, 0, 0)


from jaxgw.likelihood.detector_projection import construct_arm, detector_tensor, antenna_response


H1_lat = 46 + 27. / 60 + 18.528 / 3600
H1_long = -(119 + 24. / 60 + 27.5657 / 3600)
H1_xarm_azimuth = 125.9994
H1_yarm_azimuth = 215.9994
H1_xarm_tilt = -6.195e-4
H1_yarm_tilt = 1.25e-5

L1_lat = 30 + 33. / 60 + 46.4196 / 3600
L1_long = -(90 + 46. / 60 + 27.2654 / 3600)
L1_xarm_azimuth = 197.7165
L1_yarm_azimuth = 287.7165
L1_xarm_tilt = 0
L1_yarm_tilt = 0

H1_arm1 = construct_arm(H1_long, H1_lat, H1_xarm_tilt, H1_xarm_azimuth)
H1_arm2 = construct_arm(H1_long, H1_lat, H1_yarm_tilt, H1_yarm_azimuth)

L1_arm1 = construct_arm(L1_long, L1_lat, L1_xarm_tilt, L1_xarm_azimuth)
L1_arm2 = construct_arm(L1_long, L1_lat, L1_yarm_tilt, L1_yarm_azimuth)

H1 = detector_tensor(H1_arm1, H1_arm2)
L1 = detector_tensor(L1_arm1, L1_arm2)

H1_proj = antenna_response(H1, 1, 1, 0, 1,'plus')



import wave
import numpy as np
from scipy.interpolate import interp1d

import jax.numpy as jnp


def max_phase_diff(f, f_low, f_high, chi=1):
    gamma = np.arange(-5,6,1)/3.
    f = np.repeat(f[:,None],len(gamma),axis=1)
    f_star = np.repeat(f_low, len(gamma))
    f_star[gamma >= 0] = f_high
    return 2*np.pi*chi*np.sum((f/f_star)**gamma*np.sign(gamma),axis=1)


def make_binning_scheme(freqs, n_bins, chi=1):
    phase_diff_array = max_phase_diff(freqs,freqs[0],freqs[-1],chi=1)
    bin_f = interp1d(phase_diff_array, freqs)
    f_bins = np.array([])
    for i in np.linspace(phase_diff_array[0], phase_diff_array[-1], n_bins):
        f_bins = np.append(f_bins,bin_f(i))
    f_bins_center = (f_bins[:-1] + f_bins[1:])/2
    return f_bins, f_bins_center

def compute_coefficients(data, h_ref, psd, freqs, f_bins, f_bins_center):
    A0_array = []
    A1_array = []
    B0_array = []
    B1_array = []

    df = freqs[1] - freqs[0]
    data_prod = np.array(data*h_ref.conj())
    self_prod = np.array(h_ref*h_ref.conj())
    for i in range(len(f_bins)-1):
        f_index = np.where((freqs >= f_bins[i]) & (freqs < f_bins[i+1]))[0]
        A0_array.append(4*np.sum(data_prod[f_index]/psd[f_index]*df))
        A1_array.append(4*np.sum(data_prod[f_index]/psd[f_index]*df*(freqs[f_index]-f_bins_center[i])))
        B0_array.append(4*np.sum(self_prod[f_index]/psd[f_index]*df))
        B1_array.append(4*np.sum(self_prod[f_index]/psd[f_index]*df*(freqs[f_index]-f_bins_center[i])))

    A0_array = jnp.array(A0_array)
    A1_array = jnp.array(A1_array)
    B0_array = jnp.array(B0_array)
    B1_array = jnp.array(B1_array)
    return A0_array, A1_array, B0_array, B1_array

def make_heterodyne_likelihood(data, h_function, ref_theta, psd, freqs, n_bins=101):
    f_bins, f_bins_center = make_binning_scheme(freqs, n_bins)
    h_ref = h_function(freqs, ref_theta)
    h_ref_low = h_function(f_bins[:-1], ref_theta)
    h_ref_bincenter = h_function(f_bins_center, ref_theta)

    A0, A1, B0, B1 = compute_coefficients(data, h_ref, psd, freqs, f_bins, f_bins_center)

    def heterodyne_likelihood(params):
        waveform_low = h_function(f_bins[:-1], params)
        waveform_center = h_function(f_bins_center, params)
        
        r0 = waveform_center/h_ref_bincenter
        r1 = (waveform_low/h_ref_low - r0)/(f_bins[:-1]-f_bins_center)

        match_filter_SNR = jnp.sum(A0*r0.conj() + A1*r1.conj())
        optimal_SNR = jnp.sum(B0*jnp.abs(r0)**2 + 2*B1*(r0*r1.conj()).real)

        return (match_filter_SNR - optimal_SNR/2).real

    return heterodyne_likelihood

def make_heterodyne_likelihood_mutliple_detector(data_list, psd_list, respose_list, h_function, ref_theta, freqs, n_bins=101):

    num_detector = len(data_list)

    f_bins, f_bins_center = make_binning_scheme(freqs, n_bins)
    ra, dec = ref_theta[9], ref_theta[10]
    h_ref = []
    h_ref_low = []
    h_ref_bincenter = []
    for i in range(num_detector):
        raw_hp, raw_hc = h_function(freqs, ref_theta[:9])
        h_ref.append(respose_list[i](freqs, raw_hp, raw_hc, ra, dec, ref_theta[5],ref_theta[8]))
        raw_hp, raw_hc = h_function(f_bins[:-1], ref_theta[:9])
        h_ref_low.append(respose_list[i](f_bins[:-1], raw_hp, raw_hc, ra, dec, ref_theta[5], ref_theta[8]))
        raw_hp, raw_hc = h_function(f_bins_center, ref_theta[:9])
        h_ref_bincenter.append(respose_list[i](f_bins_center, raw_hp, raw_hc, ra, dec, ref_theta[5],ref_theta[8]))
    
    A0_array = []
    A1_array = []
    B0_array = []
    B1_array = []

    for i in range(num_detector):
        A0, A1, B0, B1 = compute_coefficients(data_list[i], h_ref[i], psd_list[i], freqs, f_bins, f_bins_center)
        A0_array.append(A0)
        A1_array.append(A1)
        B0_array.append(B0)
        B1_array.append(B1)
        
        
    def hetrodyne_likelihood(params):
        ra, dec = params[9], params[10]
        output_SNR = 0

        raw_hp_edge, raw_hc_edge = h_function(f_bins[:-1], params[:9])
        raw_hp_center, raw_hc_center = h_function(f_bins_center, params[:9])

        for i in range(num_detector):
            waveform_low = respose_list[i](f_bins[:-1], raw_hp_edge, raw_hc_edge, ra, dec, params[5], params[8])
            waveform_center = respose_list[i](f_bins_center, raw_hp_center, raw_hc_center, ra, dec, params[5], params[8])

            r0 = waveform_center/h_ref_bincenter[i]
            r1 = (waveform_low/h_ref_low[i] - r0)/(f_bins[:-1]-f_bins_center)

            match_filter_SNR = jnp.sum(A0_array[i]*r0.conj() + A1_array[i]*r1.conj())
            optimal_SNR = jnp.sum(B0_array[i]*jnp.abs(r0)**2 + 2*B1_array[i]*(r0*r1.conj()).real)

            output_SNR += (match_filter_SNR - optimal_SNR/2).real
        
        return output_SNR
    
    return hetrodyne_likelihood


import numpy as np
import bilby
import jax.numpy as jnp

from jax.config import config
from jax import grad
config.update("jax_enable_x64", True)

# Set the duration and sampling frequency of the data segment that we're
# going to inject the signal into
duration = 4.
sampling_frequency = 2048.
minimum_frequency = 20

# Specify the output directory and the name of the simulation.
outdir = 'outdir'
label = 'fast_tutorial'
bilby.core.utils.setup_logger(outdir=outdir, label=label)

# Set up a random seed for result reproducibility.  This is optional!
np.random.seed(88170235)

# We are going to inject a binary black hole waveform.  We first establish a
# dictionary of parameters that includes all of the different waveform
# parameters, including masses of the two black holes (mass_1, mass_2),
# spins of both black holes (a, tilt, phi), etc.
injection_parameters = dict(
    mass_1=36., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.5, tilt_2=1.0,
    phi_12=1.7, phi_jl=0.3, luminosity_distance=2000., theta_jn=0.4, psi=2.659,
    phase=1.3, geocent_time=1126259642.413, ra=1.375, dec=-1.2108)

# Fixed arguments passed into the source model
waveform_arguments = dict(waveform_approximant='IMRPhenomPv2',
                          reference_frequency=50.,
                          minimum_frequency=minimum_frequency)

# Create the waveform_generator using a LAL BinaryBlackHole source function
waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration, sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
    parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
    waveform_arguments=waveform_arguments)

# Set up interferometers.  In this case we'll use two interferometers
# (LIGO-Hanford (H1), LIGO-Livingston (L1). These default to their design
# sensitivity
ifos = bilby.gw.detector.InterferometerList(['H1'])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency, duration=duration,
    start_time=injection_parameters['geocent_time'] - 3)
ifos.inject_signal(waveform_generator=waveform_generator,
                   parameters=injection_parameters)

# Initialise the likelihood by passing in the interferometer data (ifos) and
# the waveform generator
likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator)

likelihood.parameters = injection_parameters
snr_bilby = likelihood.calculate_snrs(waveform_generator.frequency_domain_strain(),ifos[0]).optimal_snr_squared

##############################################
# Jax section
##############################################


waveform = waveform_generator.frequency_domain_strain()
waveform_frequency = waveform_generator.frequency_array

psd = ifos[0].power_spectral_density_array
psd_frequency = ifos[0].frequency_array

waveform_frequency = waveform_frequency[jnp.isfinite(psd)]
psd_frequency = psd_frequency[jnp.isfinite(psd)]
psd = psd[jnp.isfinite(psd)]

from jaxgw.likelihood.detector_projection import construct_arm, detector_tensor, antenna_response, get_detector_response
from jaxgw.likelihood.utils import inner_product
from jaxgw.waveform.TaylorF2 import TaylorF2

waveform = TaylorF2(waveform_frequency, injection_parameters)
H1_lat = 46 + 27. / 60 + 18.528 / 3600
H1_long = -(119 + 24. / 60 + 27.5657 / 3600)
H1_xarm_azimuth = 125.9994
H1_yarm_azimuth = 215.9994
H1_xarm_tilt = -6.195e-4
H1_yarm_tilt = 1.25e-5

H1_arm1 = construct_arm(H1_long, H1_lat, H1_xarm_tilt, H1_xarm_azimuth)
H1_arm2 = construct_arm(H1_long, H1_lat, H1_yarm_tilt, H1_yarm_azimuth)

H1 = detector_tensor(H1_arm1, H1_arm2)

strain = get_detector_response(waveform,injection_parameters,H1)
jaxgw_snr = inner_product(strain, strain, waveform_frequency, psd, psd_frequency)
d_jaxgw_snr = grad(inner_product)(strain, strain, waveform_frequency, psd, psd_frequency)

def jax_likelihood(params, data, data_f, PSD, PSD_f):
	waveform = TaylorF2(data_f, params)
	waveform = get_detector_response(waveform, params, H1)
	output = inner_product(waveform, data, data_f, PSD, PSD_f)
	return output

dlikelihood = grad(jax_likelihood)(injection_parameters, strain, waveform_frequency, psd, psd_frequency)



from astropy.constants import c,au,G,pc
from astropy.units import year as yr
from astropy.cosmology import WMAP9 as cosmo

Msun = 4.9255e-6
year = (1*yr).cgs.value
Mpc = 1e6*pc.value/c.value
euler_gamma = 0.577215664901532860606512090082
MR_sun = 1.476625061404649406193430731479084713e3
speed_of_light = 299792458.0

